@@ -11,6 +11,8 @@ from . import packet
 from . import payload
 from . import socket
 
+default_logger = logging.getLogger('engineio')
+
 
 class Server(object):
     """An Engine.IO server.
@@ -18,11 +20,14 @@ class Server(object):
     This class implements a fully compliant Engine.IO web server with support
     for websocket and long-polling transports.
 
-    :param async_mode: The library used for asynchronous operations. Valid
-                       options are "threading", "eventlet" and "gevent". If
-                       this argument is not given, "eventlet" is tried first,
-                       then "gevent", and finally "threading". The websocket
-                       transport is only supported in "eventlet" mode.
+    :param async_mode: The asynchronous model to use. See the Deployment
+                       section in the documentation for a description of the
+                       available options. Valid async modes are "threading",
+                       "eventlet", "gevent" and "gevent_uwsgi". If this
+                       argument is not given, "eventlet" is tried first, then
+                       "gevent_uwsgi", then "gevent", and finally "threading".
+                       The first async mode that has all its dependencies
+                       installed is then one that is chosen.
     :param ping_timeout: The time in seconds that the client waits for the
                          server to respond before disconnecting.
     :param ping_interval: The interval in seconds at which the client pings
@@ -47,6 +52,9 @@ class Server(object):
                  packets. Custom json modules must have ``dumps`` and ``loads``
                  functions that are compatible with the standard library
                  versions.
+    :param async_handlers: If set to ``True``, run message event handlers in
+                           non-blocking threads. To run handlers synchronously,
+                           set to ``False``. The default is ``True``.
     :param kwargs: Reserved for future extensions, any additional parameters
                    given as keyword arguments will be silently ignored.
     """
@@ -57,7 +65,8 @@ class Server(object):
                  max_http_buffer_size=100000000, allow_upgrades=True,
                  http_compression=True, compression_threshold=1024,
                  cookie='io', cors_allowed_origins=None,
-                 cors_credentials=True, logger=False, json=None, **kwargs):
+                 cors_credentials=True, logger=False, json=None,
+                 async_handlers=True, **kwargs):
         self.ping_timeout = ping_timeout
         self.ping_interval = ping_interval
         self.max_http_buffer_size = max_http_buffer_size
@@ -67,6 +76,7 @@ class Server(object):
         self.cookie = cookie
         self.cors_allowed_origins = cors_allowed_origins
         self.cors_credentials = cors_credentials
+        self.async_handlers = async_handlers
         self.sockets = {}
         self.handlers = {}
         if json is not None:
@@ -74,7 +84,7 @@ class Server(object):
         if not isinstance(logger, bool):
             self.logger = logger
         else:
-            self.logger = logging.getLogger('engineio')
+            self.logger = default_logger
             if not logging.root.handlers and \
                     self.logger.level == logging.NOTSET:
                 if logger:
@@ -83,7 +93,7 @@ class Server(object):
                     self.logger.setLevel(logging.ERROR)
                 self.logger.addHandler(logging.StreamHandler())
         if async_mode is None:
-            modes = ['eventlet', 'gevent', 'threading']
+            modes = ['eventlet', 'gevent_uwsgi', 'gevent', 'threading']
         else:
             modes = [async_mode]
         self.async = None
@@ -154,8 +164,13 @@ class Server(object):
                        (Python 3) are sent as text, and str (Python 2) and
                        bytes (Python 3) are sent as binary.
         """
-        self._get_socket(sid).send(packet.Packet(packet.MESSAGE, data=data,
-                                                 binary=binary))
+        try:
+            socket = self._get_socket(sid)
+        except KeyError:
+            # the socket is not available
+            self.logger.warning('Cannot send to sid %s', sid)
+            return
+        socket.send(packet.Packet(packet.MESSAGE, data=data, binary=binary))
 
     def disconnect(self, sid=None):
         """Disconnect a client.
@@ -285,6 +300,16 @@ class Server(object):
         th.start()
         return th  # pragma: no cover
 
+    def sleep(self, seconds=0):
+        """Sleep for the requested amount of time using the appropriate async
+        model.
+
+        This is a utility function that applications can use to put a task to
+        sleep without having to worry about using the correct call for the
+        selected async mode.
+        """
+        return self.async['sleep'](seconds)
+
     def _generate_id(self):
         """Generate a unique session id."""
         return uuid.uuid4().hex
@@ -302,7 +327,7 @@ class Server(object):
                           'pingInterval': int(self.ping_interval * 1000)})
         s.send(pkt)
 
-        if self._trigger_event('connect', sid, environ) is False:
+        if self._trigger_event('connect', sid, environ, async=False) is False:
             self.logger.warning('Application rejected connection')
             del self.sockets[sid]
             return self._unauthorized()
@@ -324,10 +349,14 @@ class Server(object):
             return []
         return ['websocket']
 
-    def _trigger_event(self, event, *args):
+    def _trigger_event(self, event, *args, **kwargs):
         """Invoke an event handler."""
+        async = kwargs.pop('async', False)
         if event in self.handlers:
-            return self.handlers[event](*args)
+            if async:
+                return self.start_background_task(self.handlers[event], *args)
+            else:
+                return self.handlers[event](*args)
 
     def _get_socket(self, sid):
         """Return the socket object for a given session."""
